# -*- python -*-
# $RCSfile: solvermenu.py,v $
# $Revision: 1.109 $
# $Author: vrc $
# $Date: 2007-09-26 22:03:04 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.SWIG.common import config
from ooflib.SWIG.common import ooferror
from ooflib.SWIG.common import switchboard
from ooflib.SWIG.common.IO import progressbar
from ooflib.SWIG.engine import masterelement
from ooflib.SWIG.engine import preconditioner
from ooflib.SWIG.engine import solver
from ooflib.SWIG.engine import solverdriver
from ooflib.common import debug
from ooflib.common import thread_enable
from ooflib.common import utils
from ooflib.common.IO import automatic
from ooflib.common.IO import mainmenu
from ooflib.common.IO import oofmenu
from ooflib.common.IO import parameter
from ooflib.common.IO import whoville
from ooflib.common.IO import xmlmenudump
from ooflib.engine import fieldinit
from ooflib.engine import skeletoncontext
from ooflib.engine import timedrivers
from ooflib.common import parallel_enable
from types import *
import ooflib.engine.mesh



solvermenu = mainmenu.OOF.addItem(oofmenu.OOFMenuItem(
    'Solver',
    help="Commands for solving Equations on Meshes.",
    cli_only=1))

if config.devel()>=1:
    def _solve(menuitem, subproblem, matrixdriver, timedriver):
        solverProgressBar = menuitem.getProgressBar("continuous") ## actually ran by the worker
        subpcontext = ooflib.engine.subproblemcontext.subproblems[subproblem]
        if subpcontext.getObject().n_active_eqns() == 0:
            raise ooferror.ErrSetupError("There are no active equations!")
        subpcontext.reserve()
        try:
            timedriver.apply(matrixdriver, subpcontext)
            ## solved flag is set even if solution did NOT
            ## converge in order to show how far the solution
            ## went.
            subpcontext.solved()

        finally:
            subpcontext.cancel_reservation()

    solvermenu.addItem(oofmenu.OOFMenuItem(
        'Solve',
        callback=_solve,
        threadable= oofmenu.THREADABLE,
        params=[whoville.WhoParameter('subproblem',
                                      ooflib.engine.subproblemcontext.subproblems,
                                      tip=parameter.emptyTipString),
                parameter.RegisteredParameter('matrixdriver', solverdriver.Driver,
                                              tip=parameter.emptyTipString),
                parameter.RegisteredParameter('timedriver', timedrivers.TimeDriver)
                ],
        help="Solve a problem",
        discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/solve.xml')
        ))



else:                                   # not devel
    def _solve(menuitem, subproblem, mesh, solver):
        if mesh is not None:
            if subproblem is not None:
                raise ooferror.ErrUserError(
                    "You must provide 'subproblem' or 'mesh', but not both!")
            subproblem = mesh + ':' + ooflib.engine.mesh.defaultSubProblemName
        subpcontext = ooflib.engine.subproblemcontext.subproblems[subproblem]
        if subpcontext.getObject().n_active_eqns() == 0:
            raise ooferror.ErrSetupError("There are no active equations!")
        if parallel_enable.enabled():
            from ooflib.engine.IO import solvermenuIPC
            solvermenuIPC.ipcsolvermenu.Solve(subproblem=subproblem,
                                              solver=solver)
            return
        solverProgressBar = menuitem.getProgressBar("continuous") ## actually ran by the worker
        subpcontext.reserve()
        timedriver = timedrivers.Null()
        try:
            try:
                # null timedriver.apply just calls solverdriver.apply
                timedriver.apply(solver, subpcontext)
            except ooferror.ErrProcessAborted:
                pass
            ## solved flag is set even if solution did NOT
            ## converge in order to show how far the solution
            ## went.
            subpcontext.solved()

        finally:
            subpcontext.cancel_reservation()
    solvermenu.addItem(oofmenu.OOFMenuItem(
        'Solve',
        callback=_solve,
        threadable= oofmenu.THREADABLE,
        params=[whoville.WhoParameter('subproblem',
                                      ooflib.engine.subproblemcontext.subproblems,
                                      tip=parameter.emptyTipString),
                whoville.WhoParameter('mesh',
                                      ooflib.engine.mesh.meshes,
                                      tip="Deprecated.  Use subproblem instead."),
                parameter.RegisteredParameter('solver', solverdriver.Driver,
                                              tip=parameter.emptyTipString)
                ],
        help="Solve a problem",
        discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/solve.xml')
        ))
    
#######################################

def _reduce_integration(menuitem, degree):
    masterelement.cvar.integration_reduction = degree

mainmenu.settingsmenu.addItem(oofmenu.OOFMenuItem(
    "Reduce_Integration",
    callback=_reduce_integration,
    params=[parameter.IntParameter("degree", 0,
                         tip="degree by which to reduce Gaussian integration")],
    help="Reduce all Gaussian integration orders by the specified degree.",
    discussion="""<para>

    Use a <emphasis>reduced integration</emphasis> scheme to do all
    element integrals.  If <varname>degree</varname> is nonzero, when
    doing an integral of a polynomial of order <varname>D</varname>
    each &mesh; element will use gauss points appropriate for an
    integral of order
    <varname>D</varname>&minus;<varname>degree</varname> instead.
    
    </para>"""
    ))

#######################################

## TODO: Subproblem-specific reset?

def _reset_mesh(menuitem, mesh):
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    try:
        femesh = meshcontext.getObject()
        ## first set time indices back to zero
        femesh.set_time(0)
        ## next set active fields back to zero
        resetObj = fieldinit.ResetFieldInit()
        for field in femesh.all_active_subproblem_fields():
            resetObj.apply(femesh, field)
        switchboard.notify("field initialized")
        ## then apply any initializers that had been
        ## applied
        initializers_dict = meshcontext.initializers
        for field in initializers_dict.keys():
            initializer = initializers_dict[field]
            try:
                initializer.apply(femesh, field)
            except:
                reporter.warning("could not re-apply initializer: ",  initializer)
        ## finally notify that solution is not current and
        ## redraw
        meshcontext.changed()
        switchboard.notify("redraw")
    finally:
        meshcontext.cancel_reservation()

if config.devel()>=1:
    solvermenu.addItem(oofmenu.OOFMenuItem(
        'Reset',
        callback=_reset_mesh,
        threadable= oofmenu.THREADABLE,
        params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                      tip=parameter.emptyTipString),

                ],
        help="Reset the fields on the mesh to their initial values"
        ))

