# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

'''
This module contains the Curve class and several routines to create
test curves, to compute the curvature of a given Curve object
and to test the accuracy of the curvature computation.

The following are used to create test curves:
   constructTestCurve( testId, nPts=8, perturb=0.0,
                       param1=0.0, param2=0.0 )
   constructGeneralizedUnitCircle( n, p )
   constructNephroid( n )
   constructFlowerCurve( a, w, n )
   perturbCurve( curve, perturb )

The following are used to compute curvature:
   computeCurvatureVariationally( curve )
   computeCurvatureLocally( curve )
   computeCurvatureFrom3Pts( pt1, pt2, pt3 )
   pointDistSqr( pt1, pt2 )
   pointDist( pt1, pt2 )

The following are used to compute the error in curvature computation:
   curvatureError( curve, computedCurvature, exactCurvature )
   curvatureErrorConvergenceTest( testCurve, N1, N2, perturb=0.0,
                                  param1=0.0, param2=0.0 )
'''

from numpy import *
from tridiag import *


'''
The following classes BdCoord, BdNode, BdEdge are defined mostly
for test purposes. They follow the Coord, Node, Edge classes in OOF
with minor modifications
'''


class BdCoord:
    'The coordinate of a 2D boundary point with the attributes x,y'

    def __init__(self,x,y):
        self.x = x;
        self.y = y;

    def __repr__(self):
        return "(%.2f,%.2f)" % (self.x, self.y)



class BdNode:
    '''
    A 2D node whose coordinate is given by Bdnode.coord of type BdCoord.
    Its coordinate can be accessed by BdNode.position()
    '''

    def __init__(self,coord):
        self.coord = coord

    def __repr__(self):
        return "(%.2f,%.2f)" % (self.coord.x, self.coord.y)

    def position(self):
        'returns the coordinate (of type BdCoord) of the BdNode object'
        return self.coord



class BdEdge:
    '''
    A 2D boundary edge defined by ( BdEdge.startNode, BdEdge.startNode ).
    The start and end nodes can be accessed by BdEdge.startnode(),.endnode()
    respectively.
    '''

    def __init__(self,startNode,endNode):
        self.startNode = startNode
        self.endNode   = endNode

    def __repr__(self):
        edgeRepr = "(%.2f,%.2f) - (%.2f,%.2f)" % \
                   ( self.startNode.position().x, self.startNode.position().y, \
                     self.endNode.position().x,   self.endNode.position().y )
        return edgeRepr

    def startnode(self):
        return self.startNode

    def endnode(self):
        return self.endNode



class Curve:
    '''
    The Curve class used to represent a 2D curve.
    It can be constructed with an edge list (of BdEdge objects).
    The following methods are available:
    To traverse the edge list
       rewind()
       edge()
       increment()
       end()
    To access geometry and perform computations
       normals()
       massMatrix()
       stiffnessMatrix()
       pwconstWeightedMassMatrix()
       normalMatrix()
    '''

    __edgeIndex = 0
    __mass  = None
    __stiff = None
    __normalMtx1 = None
    __normalMtx2 = None
    __normals = None

    def __init__(self,edgeList):
        self.edges = []
        self.nodes = []
        for edge in edgeList:
            startPt = BdCoord( edge[0][0], edge[0][1] )
            endPt   = BdCoord( edge[1][0], edge[1][1] )
            bdEdge  = BdEdge( BdNode(startPt), BdNode(endPt) )
            self.edges.append( bdEdge )
            self.nodes.append( startPt )

    def __repr__(self):
        strRepr = ""
        for edge in self.edges:
            strRepr += str(edge) + "\n"
        strRepr = strRepr.rstrip("\n")
        return strRepr

    def edge(self):
        'returns the current edge in a traversal of the edge list)'
        try:
            return self.edges[self.__edgeIndex]
        except IndexError:
            return None

    def rewind(self):
        'rewinds the edge list in traversal'
        self.__edgeIndex = 0

    def increment(self):
        'increment in edge list traversal'
        if self.__edgeIndex < len(self.edges):
            self.__edgeIndex += 1

    def end(self):
        'return if it is end of edge list in a traversal'
        return self.__edgeIndex == len(self.edges)

    def normals(self):
        'returns the normals of the edges'

        if self.__normals != None:
            return self.__normals

        N = len( self.nodes )
        self.__normals = []
        for i in range(N):
            diffX = self.nodes[(i+1) % N].x - self.nodes[i].x
            diffY = self.nodes[(i+1) % N].y - self.nodes[i].y
            diffSize = sqrt( diffX**2 + diffY**2 )
            normalX =  diffY / diffSize
            normalY = -diffX / diffSize
            self.__normals.append( (normalX, normalY) )

        return self.__normals


    def massMatrix(self):
        '''
        Returns the mass matrix

           M_{ij} = \int_\Gamma \phi_i \phi_j d\Gamma

        of the current curve \Gamma. {\phi_i} are piecewise linear
        test functions.
        '''

        if self.__mass != None: return self.__mass

        N = len( self.nodes )
        a = ones(N) / 6;
        b = ones(N) / 3;
        c = ones(N) / 6;

        for i in range( N ):
            d0 = pointDist( self.nodes[i-1], self.nodes[i] )
            d1 = pointDist( self.nodes[i],   self.nodes[(i+1) % N] )
            a[i] *= d0
            b[i] *= d0 + d1
            c[i] *= d1

        self.__mass = TridiagMatrix(a,b,c)

        return self.__mass

    def stiffnessMatrix(self):
        '''
        Returns the stiffness matrix

           A_{ij} = \int_\Gamma D\phi_i D\phi_j d\Gamma

        of the current curve \Gamma. {\phi_i} are piecewise linear
        test functions and D\phi_i denotes the tangential derivative
        of \phi_i.
        '''

        if self.__stiff != None: return self.__stiff

        N = len( self.nodes )
        a = zeros(N);
        b = zeros(N);
        c = zeros(N);

        for i in range( N ):
            d0 = pointDist( self.nodes[i-1], self.nodes[i] )
            d1 = pointDist( self.nodes[i],   self.nodes[(i+1) % N] )
            a[i] = -1./d0
            b[i] =  1./d0 + 1./d1
            c[i] = -1./d1

        self.__stiff = TridiagMatrix(a,b,c)

        return self.__stiff


    def pwconstWeightedMassMatrix(self,weight):
        '''
        Returns the weighted mass matrix

           M_{ij} = \int_\Gamma w(x) \phi_i \phi_j d\Gamma

        of the current curve \Gamma. {\phi_i} are piecewise linear
        test functions.
        weight is a vector that represents the weight function w(x)
        that is constant on each element.
        '''

        N = len( self.nodes )
        a = ones(N) / 6;
        b = ones(N) / 3;
        c = ones(N) / 6;

        for i in range( N ):
            d0 = pointDist( self.nodes[i-1], self.nodes[i] )
            d1 = pointDist( self.nodes[i],   self.nodes[(i+1) % N] )
            a[i] *= weight[i-1] * d0
            b[i] *= weight[i-1] * d0 + weight[i] * d1
            c[i] *= weight[i] *d1

        return TridiagMatrix(a,b,c)


    def normalMatrix(self,num):
        '''
        Returns the normal matrix num (=1,2) given by

           N_{ij} = \int_\Gamma n(x) \phi_i \phi_j d\Gamma

        of the current curve \Gamma. n(x) is the component num of outward
        unit normal vector and {\phi_i} are piecewise linear test functions.
        '''

        if num not in [1,2]:
            raise Exception, 'Number for normal matrix should be one of 1 or 2!'

        if ( num == 1 ) and ( self.__normalMtx1 != None ):
            return self.__normalMtx1

        if ( num == 2 ) and ( self.__normalMtx2 != None ):
            return self.__normalMtx2

        normals = self.normals()
        normalX = []
        normalY = []

        for normal in normals:
            normalX.append( normal[0] )
            normalY.append( normal[1] )

        self.__normalMtx1 = self.pwconstWeightedMassMatrix( normalX )
        self.__normalMtx2 = self.pwconstWeightedMassMatrix( normalY )

        if num == 1:
            return self.__normalMtx1
        else:
            return self.__normalMtx2



def constructGeneralizedUnitCircle(n,p):

    if p == 0.0:  p = 2.0

    dt  = 2*pi/n
    t   = arange( 0, 2*pi, dt )
    c   = cos(t)
    s   = sin(t)

    nrm    = (c**p + s**p)**(1/p)
    d_nrm  = nrm**(1-p) * (c*s**(p-1) - s*c**(p-1))
    dd_nrm = (1-p) * nrm * d_nrm**2 + \
              nrm**(1-p) * ( ((p-1)*s**2 - c**2)*c**(p-2) +
                             ((p-1)*c**2 - s**2)*s**(p-2) )

    x   =  c / nrm
    y   =  s / nrm
    dx  = -y - c * d_nrm / nrm**2
    dy  =  x - s * d_nrm / nrm**2
    ddx = -x * (1 + dd_nrm/nrm) - 2*dx * d_nrm/nrm
    ddy = -y * (1 + dd_nrm/nrm) - 2*dy * d_nrm/nrm

    if p == 2.0:
        K = ones(n)  # the exact curvature values are all ones for the circle
    else:
        K = (dx*ddy - dy*ddx) / sqrt(dx**2 + dy**2)**3
        print 'Warning: the exact curvature returned is not correct!\n'

    edgeList = [];
    for i in range(n-1):
        edge = ( (x[i],y[i]), (x[i+1],y[i+1]) )
        edgeList.append( edge )

    edge = ( (x[n-1],y[n-1]), (x[0],y[0]) )
    edgeList.append( edge )

    return (Curve(edgeList), K, x, y)


def constructNephroid(n):

    dt = 2*pi/n
    t  = arange( 0, 2*pi, dt )

    x   =  3*cos(t) + cos(3*t);
    y   =  3*sin(t) + sin(3*t);

    dx  = -3*sin(t) - 3*sin(3*t);
    dy  =  3*cos(t) + 3*cos(3*t);

    ddx = -3*cos(t) - 9*cos(3*t);
    ddy = -3*sin(t) - 9*sin(3*t);

    K   = (dx*ddy - dy*ddx) / sqrt(dx**2 + dy**2)**3
    K[n/4]   = 0.5 * ( K[n/4-1]   + K[n/4+1] );
    K[3*n/4] = 0.5 * ( K[3*n/4-1] + K[3*n/4+1] );

    edgeList = [];
    for i in range(n-1):
        edge = ( (x[i],y[i]), (x[i+1],y[i+1]) )
        edgeList.append( edge )

    edge = ( (x[n-1],y[n-1]), (x[0],y[0]) )
    edgeList.append( edge )

    return (Curve(edgeList), K)


def constructFlowerCurve(a,w,n):

    dt  = 2*pi/n
    t   = arange( 0, 2*pi, dt )

    r   = 1 + a*cos(w*t)
    dr  = -a*w*sin(w*t)
    ddr = -a*w**2*cos(w*t)

    x   = r*cos(t)
    y   = r*sin(t)

    dx  = dr*cos(t) - r*sin(t)
    dy  = dr*sin(t) + r*cos(t)

    ddx = (ddr-r)*cos(t) - 2*dr*sin(t)
    ddy = (ddr-r)*sin(t) + 2*dr*cos(t)

    K   = (dx*ddy - dy*ddx) / sqrt(dx**2 + dy**2)**3

    edgeList = [];
    for i in range(n-1):
        edge = ( (x[i],y[i]), (x[i+1],y[i+1]) )
        edgeList.append( edge )

    edge = ( (x[n-1],y[n-1]), (x[0],y[0]) )
    edgeList.append( edge )

    return (Curve(edgeList), K)



def perturbCurve(curve,perturb):

    N = len( curve.nodes )

    for i in range( N ):
        edge1  = pointDist( curve.nodes[i-1], curve.nodes[i] )
        edge2  = pointDist( curve.nodes[i],   curve.nodes[(i+1) % N] )
        factor = 0.5 * perturb * (edge1 + edge2)
        factor *= (2*random.rand() - 1)
        x = curve.nodes[i].x
        y = curve.nodes[i].y
        dx = factor * abs(x) / sqrt(x**2 + y**2)
        dy = factor * abs(y) / sqrt(x**2 + y**2)
        curve.nodes[i].x += dx  # factor * (2*random.rand() - 1)
        curve.nodes[i].y += dy  # factor * (2*random.rand() - 1)


def constructTestCurve(testId,nPts=8,perturb=0.0,param1=0.0,param2=0.0):
    '''
    Constructs a test curve specified by testId, one of "circle",
    "nephroid", "flower".
    The curve is discretized by nPts number of points.
    It is possible perturb the curve by ''perturb'' amount randomly.
    For a circle, param1 is the order = 2,4,8,...
    For a flower, param1, param2 are a,w respectively in r(t) = 1 + a*cos(w*t).
    '''

    if nPts < 3:
        print 'Number of points on the curve cannot be less than three!\n'
        return None

    if testId == "square":
        edges = [ ( (0,0), (1,0) ), \
                  ( (1,0), (1,1) ), \
                  ( (1,1), (0,1) ), \
                  ( (0,1), (0,0) )  ]
        curve = Curve( edges )
        K = zeros(4);

    elif testId == "circle":
        curve,K = constructGeneralizedUnitCircle(nPts,param1)

    elif testId == "nephroid":
        curve,K = constructNephroid(nPts)

    elif testId == "flower":
        curve,K = constructFlowerCurve(param1,param2,nPts)

    else:
        print 'Specified curve unknown!\n'
        curve = None

    if perturb > 0.0: perturbCurve( curve, perturb )

    return (curve,K)



def pointDist(pt1,pt2):

    return  sqrt( (pt2.x - pt1.x)**2 + (pt2.y - pt1.y)**2 )



def pointDistSqr(pt1,pt2):

    return (pt2.x - pt1.x)**2 + (pt2.y - pt1.y)**2



def computeCurvatureFrom3Pts(pt1,pt2,pt3):

    bottom_sqr  = pointDistSqr( pt3, pt2 )
    bottom_sqr *= pointDistSqr( pt3, pt1 )
    bottom_sqr *= pointDistSqr( pt2, pt1 )

    top  = pt1.x * (pt2.y - pt3.y)
    top += pt2.x * (pt3.y - pt1.y)
    top += pt3.x * (pt1.y - pt2.y)

    return  2 * top /  sqrt( bottom_sqr )



def computeCurvatureLocally(curve):

    N = len( curve.nodes )

    K = zeros(N)
    for i in range(N):
        K[i] = computeCurvatureFrom3Pts( curve.nodes[i-1],
                                         curve.nodes[i],
                                         curve.nodes[(i+1) % N] )
    return K


def computeCurvatureVariationally(curve):

    N = len( curve.nodes )

    A  = curve.stiffnessMatrix()
    M  = curve.massMatrix()
    N1 = curve.normalMatrix(1)
    N2 = curve.normalMatrix(2)

    X1 = zeros(N);   X2 = zeros(N)

    for i in range(N):
        X1[i] = curve.nodes[i].x
        X2[i] = curve.nodes[i].y

    K1 = tridiagSolve( M, A.matvec(X1) )
    K2 = tridiagSolve( M, A.matvec(X2) )

    K = tridiagSolve( M, N1.matvec(K1) + N2.matvec(K2) )

    return K


def curvatureError(curve,computedCurvature,exactCurvature):
    '''
    Given a curve, its exact curvature and computed curvature,
    this function returns the maximum error and L^2 error
    in the curvature.
    '''

    N = len( curve.nodes )

    diffVec = abs( computedCurvature - exactCurvature )

    error = {}
    error["max"] = max(diffVec)

    edgeLength = zeros(N)
    for i in range( N ):
        edgeLength[i] = pointDist( curve.nodes[i], curve.nodes[(i+1) % N] )

    error["L2"]  = 0.5 * sum( edgeLength * diffVec**2 )
    error["L2"] += 0.5 * sum( edgeLength * diffVec[ mod(range(1,N+1),N) ]**2 )
    error["L2"]  = sqrt( error["L2"] )

    return error


def curvatureErrorConvergenceTest(testCurveId,N1,N2,perturb=0.0,param1=0.0,param2=0.0):
    '''
    Tests the convergence of the local and variational methods for curvature
    computation. The type of the curve for the convergence test is specified
    with testCurveId, one of "circle", "nephroid", "flower".
    The curve is discretized using 2^N1,...,2^N2 points.
    See the documentation for constructTestCurve for descriptions of the other
    parameters.
    The return values are two dictionaries errorLocal, errorVariational with
    fields "max", "L2". Each field stores an error vector for the computed
    curvature, corresponding to different discretization levels up to 2^N2.
    '''

    errorL={};  errorL["max"]=zeros(N2);  errorL["L2"]=zeros(N2)
    errorV={};  errorV["max"]=zeros(N2);  errorV["L2"]=zeros(N2)

    for i in range(N1,N2):

        N = 2**i
        curve,exactK = constructTestCurve( testCurveId, N, perturb, param1, param2 )

        localK = computeCurvatureLocally( curve )
        variationalK = computeCurvatureVariationally( curve )

        error = curvatureError( curve, localK, exactK )
        errorL["max"][i] = error["max"]
        errorL["L2"][i]  = error["L2"]

        error = curvatureError( curve, variationalK, exactK )
        errorV["max"][i] = error["max"]
        errorV["L2"][i]  = error["L2"]

    return (errorL,errorV)
